package com.fjhx.email.config.exception;

import cn.hutool.http.HttpStatus;
import com.fjhx.email.config.base.R;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.RestControllerAdvice;

import javax.mail.MessagingException;
import javax.servlet.http.HttpServletRequest;
import java.util.Objects;

/**
 * 异常拦截器
 */
@Slf4j
@RestControllerAdvice
public class UnifiedExceptionHandler {

    @Autowired
    private HttpServletRequest request;

    /**
     * 自定义业务异常
     */
    @ExceptionHandler(value = ServiceException.class)
    public R handleServiceException(ServiceException e) {
        return R.error(e.getErrorCode(), e.getMessage());
    }

    /**
     * 非法数据异常
     */
    @ExceptionHandler(value = IllegalArgumentException.class)
    public R handleIllegalArgumentException(IllegalArgumentException e) {
        return R.error(HttpStatus.HTTP_BAD_REQUEST, e.getMessage());
    }

    /**
     * 请求参数异常
     */
    @ExceptionHandler(HttpMessageNotReadableException.class)
    public R handleHttpMessageNotReadableException(HttpMessageNotReadableException e) {
        log.error(request.getRequestURI(), e);

        if (Objects.requireNonNull(e.getMessage()).contains("JSON parse error")) {
            return R.error(HttpStatus.HTTP_UNSUPPORTED_TYPE, "参数格式错误");
        }
        return R.error(HttpStatus.HTTP_UNSUPPORTED_TYPE, "请使用JSON方式传参");
    }

    /**
     * 请求方式异常
     */
    @ExceptionHandler(HttpRequestMethodNotSupportedException.class)
    public R handleHttpRequestMethodNotSupportedException(HttpRequestMethodNotSupportedException e) {
        String method = e.getMethod();
        log.error("{} 不支持 {} 请求", request.getRequestURI(), method);
        return R.error(HttpStatus.HTTP_BAD_METHOD, "此方法不支持" + method + "请求");
    }

    /**
     * 其他未知异常
     */
    @ExceptionHandler(value = MessagingException.class)
    public R handleException(MessagingException e) {
        log.error(request.getRequestURI(), e);
        return R.error(HttpStatus.HTTP_INTERNAL_ERROR, "邮件相关异常，异常信息：" + e.getMessage());
    }

    /**
     * 其他未知异常
     */
    @ExceptionHandler(value = Exception.class)
    public R handleException(Exception e) {
        log.error(request.getRequestURI(), e);
        return R.error(HttpStatus.HTTP_INTERNAL_ERROR, "邮件服务出现未知异常，请联系管理员");
    }

}
